#!/usr/bin/env python
"""
Module containing the tools for data generation
"""
from __future__ import division, print_function

import sys
import os

import bilby
import deepdish

from bilby_pipe.utils import logger
from bilby_pipe.main import DataDump, parse_args
from bilby_pipe.input import Input
from bilby_pipe.bilbyargparser import BilbyArgParser


def create_parser():
    """ Generate a parser for the data_generation.py script

    Additional options can be added to the returned parser beforing calling
    `parser.parse_args` to generate the arguments`

    Returns
    -------
    parser: BilbyArgParser
        A parser with all the default options already added

    """
    parser = BilbyArgParser(ignore_unknown_config_file_keys=True)
    parser.add('--ini', is_config_file=True, help='The ini-style config file')
    parser.add('--idx', type=int, help="The level A job index", default=0)
    parser.add('--cluster', type=str,
               help='The condor cluster ID', default=None)
    parser.add('--process', type=str,
               help='The condor process ID', default=None)
    parser.add('--outdir', default='.', help='Output directory')
    parser.add('--label', default='label', help='Output label')

    det_parser = parser.add_argument_group(title='Detector arguments')
    det_parser.add(
        '--detectors', action='append',
        help=('The names of detectors to include. If given in the ini file, '
              'multiple detectors are specified by `detectors=[H1, L1]`. If '
              'given at the command line, as `--detectors H1 --detectors L1`'))
    det_parser.add('--calibration', type=int, default=2,
                   help='Which calibration to use')
    det_parser.add('--duration', type=int, default=4,
                   help='The duration of data around the event to use')
    det_parser.add("--trigger-time", default=None, type=float,
                   help="The trigger time")
    det_parser.add("--sampling-frequency", default=4096, type=int)
    det_parser.add("--channel-names", default=None, nargs="*",
                   help="Channel names to use, if not provided known "
                   "channel names will be tested.")
    det_parser.add("--query-types", default=None, nargs="*", help="Query types to "
                   "use. If not provided known query types will be tested.")
    det_parser.add('--psd-duration', default=500, type=int,
                   help='Time used to generate the PSD, default is 500.')
    det_parser.add('--minimum-frequency', default=20, type=float)
    det_parser.add('--frequency-domain-source-model', default='lal_binary_black_hole',
                   type=str, help="Name of the frequency domain source model. Can be one of"
                                  "[lal_binary_black_hole, lal_binary_neutron_star,"
                                  "lal_eccentric_binary_black_hole_no_spins, sinegaussian, "
                                  "supernova, supernova_pca_model]")

    # Method specific options below here
    data_parser = parser.add_argument_group(title='Data setting methods')
    data_parser.add('--gracedb', type=str, help='Gracedb UID', default=None)
    data_parser.add('--gps-file', type=str, help='File containing GPS times')
    data_parser.add('--injection-file', type=str, default=None,
                    help='Path to an injection file')
    data_parser.add('--waveform-approximant', default='IMRPhenomPv2', type=str,
                    help="Name of the waveform approximant for injection")
    data_parser.add('--reference-frequency', default=20, type=float,
                    help="The reference frequency")
    return parser


class DataGenerationInput(Input):
    """ Handles user-input and creation of intermediate ifo list

    Parameters
    ----------
    parser: configargparse.ArgParser, optional
        The parser containing the command line / ini file inputs
    args_list: list, optional
        A list of the arguments to parse. Defauts to `sys.argv[1:]`

    """
    def __init__(self, args, unknown_args):

        logger.info('Command line arguments: {}'.format(args))
        logger.info('Unknown command line arguments: {}'.format(unknown_args))
        self.meta_data = dict(command_line_args=args,
                              unknown_command_line_args=unknown_args,
                              injection_parameters=None)
        self.ini = args.ini
        self.cluster = args.cluster
        self.process = args.process
        self.idx = args.idx
        self.detectors = args.detectors
        self.calibration = args.calibration
        self.channel_names = args.channel_names
        self.query_types = args.query_types
        self.duration = args.duration
        self.trigger_time = args.trigger_time
        self.sampling_frequency = args.sampling_frequency
        self.psd_duration = args.psd_duration
        self.minimum_frequency = args.minimum_frequency
        self.outdir = args.outdir
        self.label = args.label
        self.frequency_domain_source_model = args.frequency_domain_source_model

        self.gracedb = args.gracedb
        self.gps_file = args.gps_file

        self.waveform_approximant = args.waveform_approximant
        self.reference_frequency = args.reference_frequency
        self.injection_file = args.injection_file

    @property
    def cluster(self):
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        try:
            self._cluster = int(cluster)
        except (ValueError, TypeError):
            logger.debug('Unable to convert input `cluster` to type int')
            self._cluster = cluster

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, process):
        try:
            self._process = int(process)
        except (ValueError, TypeError):
            logger.debug('Unable to convert input `process` to type int')
            self._process = process

    @property
    def minimum_frequency(self):
        return self._minimum_frequency

    @minimum_frequency.setter
    def minimum_frequency(self, minimum_frequency):
        self._minimum_frequency = float(minimum_frequency)

    @property
    def detectors(self):
        """ A list of the detectors to search over, e.g., ['H1', 'L1'] """
        return self._detectors

    @detectors.setter
    def detectors(self, detectors):
        """ Handles various types of user input """
        if isinstance(detectors, list):
            if len(detectors) == 1:
                det_list = self._convert_string_to_list(detectors[0])
            else:
                det_list = detectors
        else:
            raise ValueError('Input `detectors` = {} not understood'
                             .format(detectors))

        det_list.sort()
        det_list = [det.upper() for det in det_list]
        self._detectors = det_list

    @property
    def gracedb(self):
        """ The gracedb of the candidate """
        return self._gracedb

    @gracedb.setter
    def gracedb(self, gracedb):
        """ Set the gracedb ID

        At setting, will load the json candidate data and path to the frame
        cache file.
        """
        if gracedb is None:
            self._gracedb = None
        else:
            logger.info("Setting gracedb id to {}".format(gracedb))
            candidate, frame_caches = bilby.gw.utils.get_gracedb(
                gracedb, self.data_directory, self.duration, self.calibration,
                self.detectors, self.query_types)
            self.meta_data['gracedb_candidate'] = candidate
            self._gracedb = gracedb
            self.trigger_time = candidate['gpstime']
            self.frame_caches = frame_caches

    def _parse_gps_file(self):
        gps_start_times = self.read_gps_file()
        gps_start_time = gps_start_times[self.idx]
        self.trigger_time = gps_start_time + self.duration / 2.0
        self.frame_caches = self.generate_frame_cache_list_from_gpstime(gps_start_time)

    def generate_frame_cache_list_from_gpstime(self, gps_start_time):
        cache_files = []
        for det in self.detectors:
            cache_files.append(bilby.gw.utils.gw_data_find(
                det, gps_start_time=gps_start_time, duration=self.duration,
                calibration=self.calibration, outdir=self.data_directory))
        return cache_files

    @property
    def frame_caches(self):
        """ A list of paths to the frame-cache files """
        try:
            return self._frame_caches
        except AttributeError:
            raise ValueError("frame_caches list is unset")

    @frame_caches.setter
    def frame_caches(self, frame_caches):
        """ Set the frame_caches, if successfull generate the interferometer list """
        if isinstance(frame_caches, list):
            self._frame_caches = frame_caches
            self._set_interferometers_from_frame_caches(frame_caches)
        elif frame_caches is None:
            self._frame_caches = None
        else:
            raise ValueError("frame_caches list must be a list")

    def _set_interferometers_from_frame_caches(self, frame_caches):
        """ Helper method to set the interferometers from a list of frame_caches

        If no channel names are supplied, an attempt is made by bilby to
        infer the correct channel name.

        Parameters
        ----------
        frame_caches: list
            A list of strings pointing to the frame cache file
        """
        interferometers = bilby.gw.detector.InterferometerList([])
        if self.channel_names is None:
            self.channel_names = [None] * len(frame_caches)
        for cache_file, channel_name in zip(frame_caches, self.channel_names):
            interferometer = bilby.gw.detector.load_data_from_cache_file(
                cache_file, self.trigger_time, self.duration,
                self.psd_duration, channel_name)
            interferometer.minimum_frequency = self.minimum_frequency
            interferometers.append(interferometer)
        self.interferometers = interferometers

    @property
    def parameter_conversion(self):
        if 'binary_neutron_star' in self.frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters
        elif 'binary_black_hole' in self.frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters
        else:
            return None

    @property
    def injection_file(self):
        return self._injection_file

    @injection_file.setter
    def injection_file(self, injection_file):
        if injection_file is None:
            logger.debug("No injection file set")
            self._injection_file = None
        elif os.path.isfile(injection_file):
            self._injection_file = os.path.abspath(injection_file)
            injection_dict = deepdish.io.load(injection_file)
            injection_df = injection_dict['injections']
            self.injection_parameters = injection_df.iloc[self.idx].to_dict()
            self.meta_data['injection_parameters'] = self.injection_parameters
            if self.trigger_time is None:
                self.trigger_time = self.injection_parameters['geocent_time']
            self._set_interferometers_from_simulation()
        else:
            raise FileNotFoundError(
                "Injection file {} not found".format(injection_file))

    def _set_interferometers_from_simulation(self):
        waveform_arguments = dict(waveform_approximant=self.waveform_approximant,
                                  reference_frequency=self.reference_frequency,
                                  minimum_frequency=self.minimum_frequency)

        waveform_generator = bilby.gw.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=self.bilby_frequency_domain_source_model,
            parameter_conversion=self.parameter_conversion,
            waveform_arguments=waveform_arguments)

        ifos = bilby.gw.detector.InterferometerList(self.detectors)
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration,
            start_time=self.trigger_time - self.duration / 2)

        ifos.inject_signal(waveform_generator=waveform_generator,
                           parameters=self.injection_parameters)

        self.interferometers = ifos

    @property
    def interferometers(self):
        """ A bilby.gw.detector.InterferometerList """
        try:
            return self._interferometers
        except AttributeError:
            raise ValueError("interferometers unset, did you provide a set-data method?")

    @interferometers.setter
    def interferometers(self, interferometers):
        self._interferometers = interferometers

    def save_interferometer_list(self):
        data_dump = DataDump(outdir=self.data_directory, label=self.label,
                             idx=self.idx,
                             trigger_time=self.trigger_time,
                             interferometers=self.interferometers,
                             meta_data=self.meta_data)
        data_dump.to_hdf5()


def main():
    args, unknown_args = parse_args(sys.argv[1:], create_parser())
    data = DataGenerationInput(args, unknown_args)
    data.save_interferometer_list()



"""
bilby_pipe
==========

bilby_pipe is a python3 tool for automating the process of running `bilby
<https://git.ligo.org/lscsoft/bilby>`_ for gravitational parameter
estimation on computing clusters.

"""

from . import main
from . import bilbyargparser
from . import utils

__version__ = utils.get_version_information()



#!/usr/bin/env python
"""
argument parser for bilby_pipe, adapted from configargparse.ArgParser.
"""

import io
import os
import re
import sys
import argparse
import configargparse

from collections import OrderedDict


class BilbyArgParser(configargparse.ArgParser):

    def parse_known_args(self, args=None, namespace=None,
                         config_file_contents=None, env_vars=os.environ):
        """Supports all the same args as the ArgumentParser.parse_args(..),
        as well as the following additional args.

        Additional Args:
            args: a list of args as in argparse, or a string (eg. "-x -y bla")
            config_file_contents: String. Used for testing.
            env_vars: Dictionary. Used for testing.
        """

        # used while parsing args to keep track of where they came from
        _COMMAND_LINE_SOURCE_KEY = "command_line"
        _ENV_VAR_SOURCE_KEY = "environment_variables"
        _CONFIG_FILE_SOURCE_KEY = "config_file"
        _DEFAULTS_SOURCE_KEY = "defaults"

        ACTION_TYPES_THAT_DONT_NEED_A_VALUE = (
            argparse._StoreTrueAction,
            argparse._StoreFalseAction,
            argparse._CountAction,
            argparse._StoreConstAction,
            argparse._AppendConstAction)

        if args is None:
            args = sys.argv[1:]
        elif isinstance(args, str):
            args = args.split()
        else:
            args = list(args)

        # normalize args by converting args like --key=value to --key value
        normalized_args = list()
        for arg in args:
            if arg and arg[0] in self.prefix_chars and '=' in arg:
                key, value = arg.split('=', 1)
                key = key.replace("_", "-")
                normalized_args.append(key)
                normalized_args.append(value)
            else:
                if arg.startswith("--"):
                    arg.replace("_", "-")
                normalized_args.append(arg)
        args = normalized_args

        for a in self._actions:
            a.is_positional_arg = not a.option_strings

        # maps a string describing the source (eg. env var) to a settings dict
        # to keep track of where values came from (used by print_values()).
        # The settings dicts for env vars and config files will then map
        # the config key to an (argparse Action obj, string value) 2-tuple.
        self._source_to_settings = OrderedDict()
        if args:
            a_v_pair = (None, list(args))  # copy args list to isolate changes
            self._source_to_settings[_COMMAND_LINE_SOURCE_KEY] = {'': a_v_pair}

        # handle auto_env_var_prefix __init__ arg by setting a.env_var as needed
        if self._auto_env_var_prefix is not None:
            for a in self._actions:
                config_file_keys = self.get_possible_config_keys(a)
                if config_file_keys:
                    if not (a.env_var or a.is_positional_arg):
                        if not (a.is_config_file_arg):
                            if not (a.is_write_out_config_file_arg):
                                if not (isinstance(a, argparse._HelpAction)):
                                    stripped_config_file_key = config_file_keys[0].strip(self.prefix_chars)
                                    a.env_var = (self._auto_env_var_prefix + stripped_config_file_key)
                                    a.env_var = a.env_var.replace('-', '_').upper()

        # add env var settings to the commandline that aren't there already
        env_var_args = []
        actions_with_env_var_values = []
        for a in self._actions:
            if not a.is_positional_arg:
                if a.env_var:
                    if a.env_var in env_vars:
                        if not configargparse.already_on_command_line(args,
                                                                      a.option_strings):
                            actions_with_env_var_values.append(a)

        for action in actions_with_env_var_values:
            key = action.env_var
            value = env_vars[key]
            # Make list-string into list.
            if action.nargs or isinstance(action, argparse._AppendAction):
                element_capture = re.match('\[(.*)\]', value)  # noqa
                if element_capture:
                    value = [val.strip() for val in element_capture.group(1).split(',') if val.strip()]
            env_var_args += self.convert_item_to_command_line_arg(
                action, key, value)

        args = args + env_var_args

        if env_var_args:
            self._source_to_settings[_ENV_VAR_SOURCE_KEY] = OrderedDict(
                [(a.env_var, (a, env_vars[a.env_var]))
                    for a in actions_with_env_var_values])

        # before parsing any config files, check if -h was specified.
        supports_help_arg = any(
            a for a in self._actions if isinstance(a, argparse._HelpAction))
        skip_config_file_parsing = supports_help_arg and (
            "-h" in args or "--help" in args)

        # prepare for reading config file(s)
        known_config_keys = dict((config_key, action) for action in self._actions
                                 for config_key in self.get_possible_config_keys(action))

        # open the config file(s)
        config_streams = []
        if config_file_contents is not None:
            stream = io.StringIO(config_file_contents)
            stream.name = "method arg"
            config_streams = [stream]
        elif not skip_config_file_parsing:
            config_streams = self._open_config_files(args)

        # parse each config file
        for stream in reversed(config_streams):
            try:
                config_items = self._config_file_parser.parse(stream)
            except configargparse.ConfigFileParserException as e:
                self.error(e)
            finally:
                if hasattr(stream, "close"):
                    stream.close()

            # add each config item to the commandline unless it's there already
            config_args = []
            for key, value in config_items.items():
                key = key.replace("_", "-")
                if key in known_config_keys:
                    action = known_config_keys[key]
                    discard_this_key = configargparse.already_on_command_line(
                        args, action.option_strings)
                else:
                    action = None
                    discard_this_key = self._ignore_unknown_config_file_keys or \
                        configargparse.already_on_command_line(
                            args,
                            [self.get_command_line_key_for_unknown_config_file_setting(key)])

                if not discard_this_key:
                    config_args += self.convert_item_to_command_line_arg(
                        action, key, value)
                    source_key = "%s|%s" % (_CONFIG_FILE_SOURCE_KEY, stream.name)
                    if source_key not in self._source_to_settings:
                        self._source_to_settings[source_key] = OrderedDict()
                    self._source_to_settings[source_key][key] = (action, value)

            args = args + config_args

        # save default settings for use by print_values()
        default_settings = OrderedDict()
        for action in self._actions:
            cares_about_default_value = (not action.is_positional_arg or
                                         action.nargs in [argparse.OPTIONAL,
                                                          argparse.ZERO_OR_MORE])
            if (configargparse.already_on_command_line(args, action.option_strings) or
                    not cares_about_default_value or
                    action.default is None or
                    action.default == argparse.SUPPRESS or
                    isinstance(action, ACTION_TYPES_THAT_DONT_NEED_A_VALUE)):
                continue
            else:
                if action.option_strings:
                    key = action.option_strings[-1]
                else:
                    key = action.dest
                default_settings[key] = (action, str(action.default))

        if default_settings:
            self._source_to_settings[_DEFAULTS_SOURCE_KEY] = default_settings

        # parse all args (including commandline, config file, and env var)
        namespace, unknown_args = argparse.ArgumentParser.parse_known_args(
            self, args=args, namespace=namespace)
        # handle any args that have is_write_out_config_file_arg set to true
        # check if the user specified this arg on the commandline
        output_file_paths = [getattr(namespace, a.dest, None) for a in self._actions
                             if getattr(a, "is_write_out_config_file_arg", False)]
        output_file_paths = [a for a in output_file_paths if a is not None]
        self.write_config_file(namespace, output_file_paths, exit_after=True)
        return namespace, unknown_args



#!/usr/bin/env python
"""
Module containing the tools for creating injection files
"""
from __future__ import division, print_function

import os

import numpy as np
import bilby

from . import utils
from .utils import logger, BilbyPipeError


class Input(object):
    """ Superclass of input handlers """

    @property
    def idx(self):
        """ The level A job index """
        return self._idx

    @idx.setter
    def idx(self, idx):
        self._idx = idx

    @property
    def known_detectors(self):
        try:
            return self._known_detectors
        except AttributeError:
            return ['H1', 'L1', 'V1']

    @known_detectors.setter
    def known_detectors(self, known_detectors):
        self._known_detectors = self._convert_detectors_input(known_detectors)

    @property
    def detectors(self):
        """ A list of the detectors to include, e.g., ['H1', 'L1'] """
        return self._detectors

    @detectors.setter
    def detectors(self, detectors):
        self._detectors = self._convert_detectors_input(detectors)
        self._check_detectors_against_known_detectors()

    def _convert_detectors_input(self, detectors):
        if isinstance(detectors, str):
            det_list = self._split_string_by_space(detectors)
        elif isinstance(detectors, list):
            if len(detectors) == 1:
                det_list = self._split_string_by_space(detectors[0])
            else:
                det_list = detectors
        else:
            raise BilbyPipeError(
                'Input `detectors` = {} not understood'.format(detectors))

        det_list.sort()
        det_list = [det.upper() for det in det_list]
        return det_list

    def _check_detectors_against_known_detectors(self):
        for element in self.detectors:
            if element not in self.known_detectors:
                raise BilbyPipeError(
                    'detectors contains "{}" not in the known '
                    'detectors list: {} '.format(
                        element, self.known_detectors))

    @staticmethod
    def _split_string_by_space(string):
        """ Converts "H1 L1" to ["H1", "L1"] """
        return string.split(' ')

    @staticmethod
    def _convert_string_to_list(string):
        """ Converts various strings to a list """
        string = string.replace(',', ' ')
        string = string.replace('[', '')
        string = string.replace(']', '')
        string = string.replace('"', '')
        string = string.replace("'", '')
        string_list = string.split()
        return string_list

    @property
    def outdir(self):
        """ The path to the directory where output will be stored """
        return self._outdir

    @outdir.setter
    def outdir(self, outdir):
        self._outdir = os.path.abspath(outdir)
        for dr in [outdir, self.submit_directory,
                   self.data_generation_log_directory,
                   self.data_analysis_log_directory, self.data_directory,
                   self.summary_log_directory, self.result_directory]:
            utils.check_directory_exists_and_if_not_mkdir(dr)

    @property
    def submit_directory(self):
        """ The path to the directory where submit output will be stored """
        return os.path.join(self._outdir, 'submit')

    @property
    def data_generation_log_directory(self):
        """ The path to the directory where log output will be stored """
        return os.path.join(self._outdir, 'log_data_generation')

    @property
    def data_analysis_log_directory(self):
        """ The path to the directory where log output will be stored """
        return os.path.join(self._outdir, 'log_data_analysis')

    @property
    def summary_log_directory(self):
        """ The path to the directory where log output will be stored """
        return os.path.join(self._outdir, 'log_results_page')

    @property
    def data_directory(self):
        """ The path to the directory where data output will be stored """
        return os.path.join(self._outdir, 'data')

    @property
    def result_directory(self):
        """ The path to the directory where result output will be stored """
        return os.path.join(self._outdir, 'result')

    @property
    def webdir(self):
        return self._webdir

    @webdir.setter
    def webdir(self, webdir):
        if webdir is None:
            self._webdir = os.path.join(self.outdir, 'results_page')
        else:
            self._webdir = webdir

    @property
    def gps_file(self):
        """ The gps file containing the list of gps times """
        return self._gps_file

    @gps_file.setter
    def gps_file(self, gps_file):
        """ Set the gps_file

        At setting, will check the file exists, read  the contents, identify
        which element to generate data for, and set the cache file
        """
        if gps_file is None:
            self._gps_file = None
            return
        elif os.path.isfile(gps_file):
            self._gps_file = os.path.abspath(gps_file)
        else:
            raise FileNotFoundError(
                "Input file gps_file={} not understood".format(gps_file))

        try:
            self._parse_gps_file()
        except AttributeError:
            logger.debug("No _parse_gps_file method present")

    def read_gps_file(self):
        gpstimes = np.loadtxt(self.gps_file, ndmin=1)
        return gpstimes

    @property
    def bilby_frequency_domain_source_model(self):
        """ The bilby function to pass to the waveform_generator """
        if self.frequency_domain_source_model in bilby.gw.source.__dict__.keys():
            return bilby.gw.source.__dict__[self._frequency_domain_source_model]
        else:
            raise BilbyPipeError(
                "No source model {} found.".format(
                    self._frequency_domain_source_model))

    @property
    def frequency_domain_source_model(self):
        """ String of which frequency domain source model to use """
        return self._frequency_domain_source_model

    @frequency_domain_source_model.setter
    def frequency_domain_source_model(self, frequency_domain_source_model):
        self._frequency_domain_source_model = frequency_domain_source_model



#!/usr/bin/env python
"""
Script to analyse the stored data
"""
from __future__ import division, print_function

import sys
import os

import numpy as np
import bilby

from bilby_pipe.utils import logger, BilbyPipeError
from bilby_pipe.main import DataDump, parse_args
from bilby_pipe.input import Input
from bilby_pipe.bilbyargparser import BilbyArgParser


def create_parser():
    """ Generate a parser for the data_analysis.py script

    Additional options can be added to the returned parser before calling
    `parser.parse_args` to generate the arguments`

    Returns
    -------
    parser: BilbyArgParser
        A parser with all the default options already added

    """
    parser = BilbyArgParser(ignore_unknown_config_file_keys=True)
    parser.add('--ini', is_config_file=True, help='The ini-style config file')
    parser.add('--idx', type=int, help="The level A job index", default=0)
    parser.add('--cluster', type=str,
               help='The condor cluster ID', default=None)
    parser.add('--process', type=str,
               help='The condor process ID', default=None)
    parser.add(
        '--detectors', action='append',
        help=('The names of detectors to analyse. If given in the ini file, '
              'multiple detectors are specified by `detectors=[H1, L1]`. If '
              'given at the command line, as `--detectors H1 --detectors L1`'))
    parser.add("--prior-file", default=None, help="The prior file")
    parser.add("--deltaT", type=float, default=0.1,
               help=("The symmetric width (in s) around the trigger time to"
                     " search over the coalesence time"))
    parser.add('--reference-frequency', default=20, type=float,
               help="The reference frequency")
    parser.add('--waveform-approximant', default='IMRPhenomPv2', type=str,
               help="The name of the waveform approximant")
    parser.add('--default-prior', default='BBHPriorDict', type=str,
               help="The name of the prior set to base the prior on. Can be one of"
                    "[PriorDict, BBHPriorDict, BNSPriorDict, CalibrationPriorDict]")
    parser.add('--frequency-domain-source-model', default='lal_binary_black_hole',
               type=str, help="Name of the frequency domain source model. Can be one of"
                              "[lal_binary_black_hole, lal_binary_neutron_star,"
                              "lal_eccentric_binary_black_hole_no_spins, sinegaussian, "
                              "supernova, supernova_pca_model]")
    parser.add(
        '--distance-marginalization', action='store_true', default=False,
        help='Boolean. If true, use a distance-marginalized likelihood')
    parser.add(
        '--phase-marginalization', action='store_true', default=False,
        help='Boolean. If true, use a phase-marginalized likelihood')
    parser.add(
        '--time-marginalization', action='store_true', default=False,
        help='Boolean. If true, use a time-marginalized likelihood')
    parser.add('--sampler', default=None)
    parser.add('--sampler-kwargs', default=None)
    parser.add('--outdir', default='.', help='Output directory')
    parser.add('--label', default='label', help='Output label')
    parser.add('--data-label', default=None, help='Label used for the data dump', required=True)
    parser.add('--sampling-seed', default=None, type=int, help='Random sampling seed')
    parser.add('--create-output', action='store_true',
               help='If true, create plots')
    return parser


class DataAnalysisInput(Input):
    """ Handles user-input and analysis of intermediate ifo list

    Parameters
    ----------
    parser: BilbyArgParser, optional
        The parser containing the command line / ini file inputs
    args_list: list, optional
        A list of the arguments to parse. Defauts to `sys.argv[1:]`

    """

    def __init__(self, args, unknown_args):
        logger.info('Command line arguments: {}'.format(args))

        self.ini = args.ini
        self.idx = args.idx
        self.cluster = args.cluster
        self.process = args.process
        self.detectors = args.detectors
        self.prior_file = args.prior_file
        self._priors = None
        self.deltaT = args.deltaT
        self.reference_frequency = args.reference_frequency
        self.waveform_approximant = args.waveform_approximant
        self.distance_marginalization = args.distance_marginalization
        self.phase_marginalization = args.phase_marginalization
        self.time_marginalization = args.time_marginalization
        self.sampling_seed = args.sampling_seed
        self.sampler = args.sampler
        self.sampler_kwargs = args.sampler_kwargs
        self.outdir = args.outdir
        self.label = args.label
        self.data_label = args.data_label
        self.default_prior = args.default_prior
        self.frequency_domain_source_model = args.frequency_domain_source_model
        self.result = None

    @property
    def cluster(self):
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        try:
            self._cluster = int(cluster)
        except (ValueError, TypeError):
            logger.debug('Unable to convert input `cluster` to type int')
            self._cluster = cluster

    @property
    def process(self):
        return self._process

    @process.setter
    def process(self, process):
        try:
            self._process = int(process)
        except (ValueError, TypeError):
            logger.debug('Unable to convert input `process` to type int')
            self._process = process

    @property
    def reference_frequency(self):
        return self._reference_frequency

    @reference_frequency.setter
    def reference_frequency(self, reference_frequency):
        self._reference_frequency = float(reference_frequency)

    @property
    def sampling_seed(self):
        return self._samplng_seed

    @sampling_seed.setter
    def sampling_seed(self, sampling_seed):
        if sampling_seed is None:
            sampling_seed = np.random.randint(1, 1e6)
        self._samplng_seed = sampling_seed
        np.random.seed(sampling_seed)
        logger.info('Sampling seed set to {}'.format(sampling_seed))

    @property
    def sampler_kwargs(self):
        return self._sampler_kwargs

    @sampler_kwargs.setter
    def sampler_kwargs(self, sampler_kwargs):
        if sampler_kwargs is not None:
            try:
                self._sampler_kwargs = eval(sampler_kwargs)
            except (NameError, TypeError) as e:
                raise BilbyPipeError(
                    "Error {}. Unable to parse sampler_kwargs: {}"
                    .format(e, sampler_kwargs))
        else:
            self._sampler_kwargs = None

    @property
    def interferometers(self):
        try:
            return self._interferometers
        except AttributeError:
            ifos = self.data_dump.interferometers
            names = [ifo.name for ifo in ifos]
            logger.info("Found data for detectors = {}".format(names))
            ifos_to_use = [ifo for ifo in ifos if ifo.name in self.detectors]
            names_to_use = [ifo.name for ifo in ifos_to_use]
            logger.info("Using data for detectors = {}".format(names_to_use))
            self._interferometers = bilby.gw.detector.InterferometerList(ifos_to_use)
            return self._interferometers

    @property
    def meta_data(self):
        return self.data_dump.meta_data

    @property
    def trigger_time(self):
        return self.data_dump.trigger_time

    @property
    def data_dump(self):
        try:
            return self._data_dump
        except AttributeError:
            filename = os.path.join(
                self.data_directory,
                '_'.join([self.data_label, str(self.idx), 'data_dump.h5']))
            self._data_dump = DataDump.from_hdf5(filename)
            return self._data_dump

    @property
    def priors(self):
        if self._priors is None:
            if self.default_prior in bilby.core.prior.__dict__.keys():
                self._priors = bilby.core.prior.__dict__[self.default_prior](
                    filename=self.prior_file
                )
            elif self.default_prior in bilby.gw.prior.__dict__.keys():
                self._priors = bilby.gw.prior.__dict__[self.default_prior](
                    filename=self.prior_file
                )
            else:
                logger.info("No prior {} found.").format(self.default_prior)
                logger.info("Defaulting to BBHPriorDict")
                self._priors = bilby.gw.prior.BBHPriorDict(
                    filename=self.prior_file
                )
            if isinstance(self._priors, (bilby.gw.prior.BBHPriorDict, bilby.gw.prior.BNSPriorDict)):
                self._priors['geocent_time'] = bilby.core.prior.Uniform(
                    minimum=self.trigger_time - self.deltaT / 2,
                    maximum=self.trigger_time + self.deltaT / 2,
                    name='geocent_time', latex_label='$t_c$', unit='$s$')
        return self._priors

    @property
    def parameter_conversion(self):
        if 'binary_neutron_star' in self._frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters
        elif 'binary_black_hole' in self._frequency_domain_source_model:
            return bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters
        else:
            return None

    @property
    def waveform_generator(self):
        waveform_generator = bilby.gw.WaveformGenerator(
            sampling_frequency=self.interferometers.sampling_frequency,
            duration=self.interferometers.duration,
            frequency_domain_source_model=self.bilby_frequency_domain_source_model,
            parameter_conversion=self.parameter_conversion,
            start_time=self.interferometers.start_time,
            waveform_arguments=self.waveform_arguments)
        return waveform_generator

    @property
    def waveform_arguments(self):
        return dict(
            reference_frequency=self.reference_frequency,
            waveform_approximant=self.waveform_approximant,
            minimum_frequency=self.interferometers[0].minimum_frequency)  # FIXME

    @property
    def likelihood(self):
        return bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=self.priors,
            phase_marginalization=self.phase_marginalization,
            distance_marginalization=self.distance_marginalization,
            time_marginalization=self.time_marginalization)

    @property
    def parameter_generation(self):
        if 'binary_neutron_star' in self._frequency_domain_source_model:
            return bilby.gw.conversion.generate_all_bns_parameters
        elif 'binary_black_hole' in self._frequency_domain_source_model:
            return bilby.gw.conversion.generate_all_bbh_parameters
        else:
            return None

    def run_sampler(self):
        self.result = bilby.run_sampler(
            likelihood=self.likelihood, priors=self.priors,
            sampler=self.sampler, label=self.label, outdir=self.result_directory,
            conversion_function=self.parameter_generation,
            injection_parameters=self.data_dump.meta_data['injection_parameters'],
            **self.sampler_kwargs)


def main():
    args, unknown_args = parse_args(sys.argv[1:], create_parser())
    analysis = DataAnalysisInput(args, unknown_args)
    analysis.run_sampler()



#!/usr/bin/env python
"""
bilby_pipe is a command line tools for taking user input (as command line
arguments or an ini file) and creating DAG files for submitting bilby parameter
estimation jobs.
"""
import os
import sys
import shutil
import subprocess
import itertools
import pycondor
import deepdish
from collections import namedtuple

from .utils import logger, parse_args, BilbyPipeError
from . import utils
from . import create_injections
from .input import Input

from bilby_pipe.bilbyargparser import BilbyArgParser

__version__ = utils.get_version_information()

JobInput = namedtuple('level_A_job_input', 'idx meta_label kwargs')


def create_parser():
    """ Creates the BilbyArgParser for bilby_pipe """
    parser = BilbyArgParser(
        usage=__doc__, ignore_unknown_config_file_keys=True,
        allow_abbrev=False)
    parser.add(
        'ini', type=str, is_config_file=True, help='The ini file')
    parser.add(
        '--submit', action='store_true',
        help='Attempt to submit the job after the build')
    parser.add(
        '--sampler', nargs='+', default='dynesty',
        help='Sampler to use, or list of sampler to use')
    parser.add(
        '--detectors', nargs='+', default=['H1', 'L1'],
        help='The names of detectors to include {H1, L1}')
    parser.add(
        '--coherence-test', action='store_true',
        help=('Run the analysis for all detectors together and for each '
              'detector separately'))
    parser.add(
        '--label', type=str, default='LABEL', help='The output label')
    parser.add(
        '--outdir', type=str, default='.', help='The output directory')
    parser.add(
        '--create-summary', action='store_true',
        help='If true, create a summary page')
    parser.add(
        '--webdir', type=str, default=None,
        help=('Directory to store summary pages. If not given, defaults to '
              'outdir/results_page'))
    parser.add(
        '--email', type=str,
        help='Email for notifications')
    parser.add(
        '--existing-dir', type=str, default=None,
        help=('If given, add results to an directory with an an existing'
              ' summary.html file'))
    parser.add(
        '--accounting', type=str, required=True,
        help='The accounting group to use')
    parser.add(
        '--X509', type=str, default=None,
        help=('The path to the users X509 certificate file.'
              'If not given, a copy of the file at the env. variable '
              '$X509_USER_PROXY will be made in outdir and linked in '
              'the condor jobs submission'))
    parser.add(
        '-v', '--verbose', action='store_true', help='verbose')
    parser.add(
        '--version', action='version',
        version='%(prog)s {version}'.format(version=__version__))
    parser.add(
        '--local', action='store_true',
        help='Run the job locally, i.e., not through a batch submission')
    parser.add(
        '--singularity-image', type=str, default=None,
        help='Singularity image to use')

    injection_parser = parser.add_argument_group(title='Injection arguments')
    injection_parser.add(
        '--injection', action='store_true', default=False,
        help='Create data from an injection file')
    injection_parser.add(
        '--injection-file', type=str, default=None,
        help='An injection file: overrides `n-injection`.')
    injection_parser.add_arg(
        '--n-injection', type=int,
        help='The number of injections to generate by sampling from the prior')

    data_gen_pars = parser.add_argument_group(
        title='Data generation arguments')
    data_gen_pars.add(
        '--gps-file', type=str, help='File containing GPS times')
    data_gen_pars.add(
        '--gracedb', type=str, help='Gracedb UID', default=None)
    return parser


class MainInput(Input):
    """ An object to hold all the inputs to bilby_pipe

    Parameters
    ----------
    parser: BilbyArgParser, optional
        The parser containing the command line / ini file inputs
    args_list: list, optional
        A list of the arguments to parse. Defauts to `sys.argv[1:]`

    Attributes
    ----------
    ini: str
        The path to the ini file
    submit: bool
        If true, user-input to also submit the jobs
    label: str
        A label describing the job
    outdir: str
        The path to the directory where output will be stored
    create_summary: bool
        If true, create a summary page
    accounting: str
        The accounting group to use
    coherence_test: bool
        If true, run the coherence test
    detectors: list
        A list of the detectors to include, e.g., ['H1', 'L1']
    unknown_args: list
        A list of unknown command line arguments
    x509userproxy: str
        A path to the users X509 certificate used for authentication

    """

    def __init__(self, args, unknown_args):
        logger.debug('Creating new Input object')
        logger.info('Command line arguments: {}'.format(args))

        logger.debug('Known detector list = {}'.format(self.known_detectors))

        self.unknown_args = unknown_args
        self.ini = args.ini
        self.submit = args.submit
        self.singularity_image = args.singularity_image
        self.outdir = args.outdir
        self.label = args.label
        self.queue = 1
        self.create_summary = args.create_summary
        self.accounting = args.accounting
        self.sampler = args.sampler
        self.detectors = args.detectors
        self.coherence_test = args.coherence_test
        self.x509userproxy = args.X509

        self.webdir = args.webdir
        self.email = args.email
        self.existing_dir = args.existing_dir

        self.run_local = args.local

        self.gps_file = args.gps_file

        self.injection = args.injection
        self.injection_file = args.injection_file
        self.n_injection = args.n_injection

        self.gracedb = args.gracedb

        # These keys are used in the webpages summary
        self.meta_keys = ['label', 'outdir', 'ini',
                          'detectors', 'coherence_test',
                          'sampler', 'accounting']

    @property
    def ini(self):
        return self._ini

    @ini.setter
    def ini(self, ini):
        if os.path.isfile(ini) is False:
            raise BilbyPipeError('ini file is not a file')
        self._ini = os.path.abspath(ini)

    @property
    def singularity_image(self):
        return self._singularity_image

    @singularity_image.setter
    def singularity_image(self, singularity_image):
        if singularity_image is None:
            self._singularity_image = None
            self.use_singularity = False
        elif isinstance(singularity_image, str):
            self._verify_singularity(singularity_image)
            self._singularity_image = os.path.abspath(singularity_image)
            self.use_singularity = True
        else:
            raise BilbyPipeError(
                "simg={} not understood".format(singularity_image))

    def _verify_singularity(self, singularity_image):
        """ Verify the singularity image exists """
        if os.path.isfile(singularity_image) is False:
            raise FileNotFoundError(
                "singularity_image={} is not a file".format(singularity_image))

    @property
    def use_singularity(self):
        return self._use_singularity

    @use_singularity.setter
    def use_singularity(self, use_singularity):
        if isinstance(use_singularity, bool):
            logger.debug('Setting use_singularity={}'.format(use_singularity))
            self._use_singularity = use_singularity
        else:
            raise BilbyPipeError(
                "use_singularity={} not understood".format(use_singularity))

    @property
    def n_level_A_jobs(self):
        try:
            return self._n_level_A_jobs
        except AttributeError:
            logger.debug("n_level_A_jobs not set, defaulting to 1")
            return 1

    @n_level_A_jobs.setter
    def n_level_A_jobs(self, n_level_A_jobs):
        logger.debug("Setting n_level_A_jobs = {}".format(n_level_A_jobs))
        self._n_level_A_jobs = n_level_A_jobs

    @property
    def level_A_labels(self):
        try:
            return self._level_A_jobs
        except AttributeError:
            raise BilbyPipeError("level_A_labels not set")

    @level_A_labels.setter
    def level_A_labels(self, labels):
        self._level_A_jobs = labels

    @property
    def x509userproxy(self):
        """ A path to the users X509 certificate used for authentication """
        try:
            return self._x509userproxy
        except AttributeError:
            raise BilbyPipeError(
                "The X509 user proxy has not been correctly set, please check"
                " the logs")

    @x509userproxy.setter
    def x509userproxy(self, x509userproxy):
        if x509userproxy is None:
            cert_alias = 'X509_USER_PROXY'
            try:
                cert_path = os.environ[cert_alias]
                new_cert_path = os.path.join(
                    self.outdir, '.' + os.path.basename(cert_path))
                shutil.copyfile(cert_path, new_cert_path)
                self._x509userproxy = new_cert_path
            except FileNotFoundError:
                logger.warning(
                    "Environment variable X509_USER_PROXY does not point to a"
                    " file. Try running `$ ligo-proxy-init albert.einstein`")
            except KeyError:
                logger.warning(
                    "Environment variable X509_USER_PROXY not set"
                    " Try running `$ ligo-proxy-init albert.einstein`")
                self._x509userproxy = None
        elif os.path.isfile(x509userproxy):
            self._x509userproxy = x509userproxy
        else:
            raise BilbyPipeError('Input X509 not a file or not understood')

    def _parse_gps_file(self):
        gpstimes = self.read_gps_file()
        n = len(gpstimes)
        logger.info(
            "{} times found in gps_file={}, setting level A jobs"
            .format(n, self.gps_file))

        self.n_level_A_jobs = n
        self.level_A_labels = [str(x) for x in gpstimes]

    @property
    def n_injection(self):
        return self._n_injection

    @n_injection.setter
    def n_injection(self, n_injection):
        if n_injection is not None:
            logger.info(
                "n_injection={}, setting level A jobs".format(n_injection))
            self.n_level_A_jobs = n_injection
            self._n_injection = n_injection
            self.level_A_labels = ['injection_{}'.format(x) for x in range(n_injection)]
        else:
            self._n_injection = None

    @property
    def gracedb(self):
        return self._gracedb

    @gracedb.setter
    def gracedb(self, gracedb):
        if gracedb is not None:
            self._gracedb = gracedb
            self.level_A_labels = [gracedb]
        else:
            self._gracedb = None


class Dag(object):
    """ A class to handle the creation and building of a DAG

    Parameters
    ----------
    inputs: bilby_pipe.Input
        An object holding the inputs built from the command-line and ini file.

    Other parameters
    ----------------
    request_memory : str or None, optional
        Memory request to be included in submit file.
        request_disk : str or None, optional
        Disk request to be included in submit file.
    request_cpus : int or None, optional
        Number of CPUs to request in submit file.
    getenv : bool or None, optional
        Whether or not to use the current environment settings when running
        the job (default is None).
    universe : str or None, optional
        Universe execution environment to be specified in submit file
        (default is None).
    initialdir : str or None, optional
        Initial directory for relative paths (defaults to the directory was
        the job was submitted from).
    notification : str or None, optional
        E-mail notification preference (default is None).
    requirements : str or None, optional
        Additional requirements to be included in ClassAd.
    extra_lines : list or None, optional
        List of additional lines to be added to submit file.
    dag : Dagman, optional
        If specified, Job will be added to dag (default is None).
    arguments : str or iterable, optional
        Arguments with which to initialize the Job list of arguments
        (default is None).
    retry : int or None, optional
        Option to specify the number of retries for all Job arguments. This
        can be superseded for arguments added via the add_arg() method.
        Note: this feature is only available to Jobs that are submitted via
        a Dagman (default is None; no retries).
    verbose : int, optional
        Level of logging verbosity option are 0-warning, 1-info,
        2-debugging (default is 0).

    Notes
    -----
        The "Other Parameters" are passed directly to
        `pycondor.Job()`. Documentation for these is taken verbatim from the
        API available at https://jrbourbeau.github.io/pycondor/api.html

    """

    def __init__(self, inputs, request_memory=None, request_disk=None,
                 request_cpus=None, getenv=True, universe='vanilla',
                 initialdir=None, notification='never', requirements=None,
                 retry=None, verbose=0):
        self.request_memory = request_memory
        self.request_disk = request_disk
        self.request_cpus = request_disk
        self.getenv = getenv
        self.universe = universe
        self.initialdir = initialdir
        self.notification = notification
        self.requirements = requirements
        self.retry = retry
        self.verbose = verbose
        self.inputs = inputs

        self.dag = pycondor.Dagman(
            name='dag_' + inputs.label,
            submit=self.inputs.submit_directory)
        self.generation_jobs = []
        self.generation_job_labels = []
        self.analysis_jobs = []
        self.summary_jobs = []
        self.results_pages = dict()
        if self.inputs.injection:
            self.check_injection()
        self.create_generation_jobs()
        self.create_analysis_jobs()
        self.create_postprocessing_jobs()
        if self.inputs.create_summary:
            self.create_summary_jobs()
        self.build_submit()

    @staticmethod
    def _get_executable_path(exe_name):
        exe = shutil.which(exe_name)
        if exe is not None:
            return exe
        else:
            raise OSError("{} not installed on this system, unable to proceed"
                          .format(exe_name))

    @property
    def generation_executable(self):
        if self.inputs.use_singularity:
            # This hard-codes the path to singularity: TODO infer this from the users env
            return '/bin/singularity'
        else:
            return self._get_executable_path('bilby_pipe_generation')

    @property
    def analysis_executable(self):
        if self.inputs.use_singularity:
            # This hard-codes the path to singularity: TODO infer this from the users env
            return '/bin/singularity'
        else:
            return self._get_executable_path('bilby_pipe_analysis')

    @property
    def summary_executable(self):
        return self._get_executable_path('summarypages.py')

    def check_injection(self):
        """ If injections are requested, create an injection file """
        default_injection_file_name = '{}/{}_injection_file.h5'.format(
            self.inputs.data_directory, self.inputs.label)
        if self.inputs.injection_file is not None:
            logger.info("Using injection file {}".format(self.inputs.injection_file))
        elif os.path.isfile(default_injection_file_name):
            # This is done to avoid overwriting the injection file
            logger.info("Using injection file {}".format(default_injection_file_name))
            self.inputs.injection_file = default_injection_file_name
        else:
            logger.info("No injection file found, generating one now")
            inj_args, inj_unknown_args = parse_args(
                sys.argv[1:], create_injections.create_parser())
            inj_inputs = create_injections.CreateInjectionInput(
                inj_args, inj_unknown_args)
            inj_inputs.create_injection_file(default_injection_file_name)
            self.inputs.injection_file = default_injection_file_name

    @property
    def generation_jobs_inputs(self):
        """ A list of dictionaries enumerating all the generation jobs

        This contains the logic of generating multiple parallel running jobs
        The keys of each dictionary should be the keyword arguments to
        `self._create_jobs()`

        """
        try:
            return self._generation_jobs_inputs
        except AttributeError:
            logger.debug("Generating list of generation jobs")
            jobs_numbers = range(self.inputs.n_level_A_jobs)
            jobs_inputs = []
            for idx in jobs_numbers:
                ji = JobInput(
                    idx=idx, meta_label=self.inputs.level_A_labels[idx],
                    kwargs=dict())
                jobs_inputs.append(ji)
            logger.debug("List of job inputs = {}".format(jobs_inputs))
            self._generation_jobs_inputs = jobs_inputs
            return jobs_inputs

    def create_generation_jobs(self):
        """ Create all the condor jobs and add them to the dag """
        for job_input in self.generation_jobs_inputs:
            self.generation_jobs.append(
                self._create_generation_job(job_input))

    def _create_generation_job(self, job_input):
        """ Create a job to generate the data """
        idx = job_input.idx
        job_name = '_'.join([self.inputs.label, 'generation', str(idx)])
        if job_input.meta_label is not None:
            job_name = '_'.join([job_name, job_input.meta_label])
        job_logs_base = os.path.join(self.inputs.data_generation_log_directory, job_name)
        submit = self.inputs.submit_directory
        extra_lines = ''
        for arg in ['error', 'log', 'output']:
            extra_lines += '\n{} = {}_$(Cluster)_$(Process).{}'.format(
                arg, job_logs_base, arg[:3])
        extra_lines += '\naccounting_group = {}'.format(self.inputs.accounting)
        extra_lines += '\nx509userproxy = {}'.format(self.inputs.x509userproxy)

        arguments = ArgumentsString()
        if self.inputs.use_singularity:
            arguments.append(
                'run --app generation {}'.format(self.inputs.singularity_image))
        arguments.add('ini', self.inputs.ini)
        arguments.add('label', job_name)
        self.generation_job_labels.append(job_name)
        arguments.add('idx', idx)
        arguments.add('cluster', '$(Cluster)')
        arguments.add('process', '$(Process)')
        if self.inputs.injection_file is not None:
            arguments.add('injection-file', self.inputs.injection_file)
        arguments.add_unknown_args(self.inputs.unknown_args)
        arguments.add_command_line_arguments()
        generation_job = pycondor.Job(
            name=job_name,
            executable=self.generation_executable,
            submit=submit,
            request_memory=self.request_memory, request_disk=self.request_disk,
            request_cpus=self.request_cpus, getenv=self.getenv,
            universe=self.universe, initialdir=self.initialdir,
            notification=self.notification, requirements=self.requirements,
            queue=self.inputs.queue, extra_lines=extra_lines, dag=self.dag,
            arguments=arguments.print(), retry=self.retry, verbose=self.verbose)
        logger.debug('Adding job: {}'.format(job_name))

        if self.inputs.run_local:
            subprocess.run(
                [self.generation_executable] + arguments.argument_list)

        return generation_job

    def create_analysis_jobs(self):
        """ Create all the condor jobs and add them to the dag """
        for job_input in self.analysis_jobs_inputs:
            self.analysis_jobs.append(self._create_analysis_job(job_input))

    @property
    def analysis_jobs_inputs(self):
        """ A list of dictionaries enumerating all the main jobs to generate
        This contains the logic of generating multiple parallel running jobs
        The keys of each dictionary should be the keyword arguments to
        `self._create_jobs()`

        """
        logger.debug("Generating list of jobs")

        # Create level B inputs
        detectors_list = []
        detectors_list.append(self.inputs.detectors)
        if self.inputs.coherence_test:
            for detector in self.inputs.detectors:
                detectors_list.append([detector])
        sampler_list = self.inputs.sampler
        level_B_prod_list = list(itertools.product(detectors_list, sampler_list))

        level_A_jobs_numbers = range(self.inputs.n_level_A_jobs)
        jobs_inputs = []
        for idx in list(level_A_jobs_numbers):
            for detectors, sampler in level_B_prod_list:
                jobs_inputs.append(
                    JobInput(idx=idx, meta_label=self.inputs.level_A_labels[idx],
                             kwargs=dict(detectors=detectors, sampler=sampler)))

        logger.debug("List of job inputs = {}".format(jobs_inputs))
        return jobs_inputs

    def _create_analysis_job(self, job_input):
        """ Create a condor job and add it to the dag

        Parameters
        ----------
        detectors: list, str
            A list of the detectors to include, e.g. `['H1', 'L1']`
        sampler: str
            The sampler to use for the job

        """
        detectors = job_input.kwargs['detectors']
        sampler = job_input.kwargs['sampler']
        idx = job_input.idx
        if not isinstance(detectors, list):
            raise BilbyPipeError("`detectors must be a list")

        job_name = '_'.join([self.inputs.label, ''.join(detectors), sampler])
        if job_input.meta_label is not None:
            job_name = '_'.join([job_name, job_input.meta_label])
        job_logs_base = os.path.join(self.inputs.data_analysis_log_directory, job_name)
        submit = self.inputs.submit_directory
        extra_lines = ''
        for arg in ['error', 'log', 'output']:
            extra_lines += '\n{} = {}_$(Cluster)_$(Process).{}'.format(
                arg, job_logs_base, arg[:3])
        extra_lines += '\naccounting_group = {}'.format(self.inputs.accounting)
        extra_lines += '\nx509userproxy = {}'.format(self.inputs.x509userproxy)

        arguments = ArgumentsString()
        if self.inputs.use_singularity:
            arguments.append(
                'run --app analysis {}'.format(self.inputs.singularity_image))
        arguments.add('ini', self.inputs.ini)
        for detector in detectors:
            arguments.add('detectors', detector)
        arguments.add('label', job_name)
        arguments.add('data-label', self.generation_job_labels[idx])
        arguments.add('idx', idx)
        arguments.add('sampler', sampler)
        arguments.add('cluster', '$(Cluster)')
        arguments.add('process', '$(Process)')
        arguments.add_unknown_args(self.inputs.unknown_args)
        arguments.add_command_line_arguments()
        job = pycondor.Job(
            name=job_name,
            executable=self.analysis_executable,
            submit=submit,
            request_memory=self.request_memory, request_disk=self.request_disk,
            request_cpus=self.request_cpus, getenv=self.getenv,
            universe=self.universe, initialdir=self.initialdir,
            notification=self.notification, requirements=self.requirements,
            queue=self.inputs.queue, extra_lines=extra_lines, dag=self.dag,
            arguments=arguments.print(), retry=self.retry, verbose=self.verbose)
        job.add_parent(self.generation_jobs[idx])
        logger.debug('Adding job: {}'.format(job_name))
        self.results_pages[job_name] = 'result/{}.html'.format(job_name)

        if self.inputs.run_local:
            subprocess.run(
                [self.analysis_executable] + arguments.argument_list)

        return job

    def create_postprocessing_jobs(self):
        """ Generate postprocessing job """
        pass

    @property
    def summary_jobs_inputs(self):
        """ A list of dictionaries enumerating all the main jobs to generate

        This contains the logic of generating multiple parallel running jobs
        The keys of each dictionary should be the keyword arguments to
        `self._create_jobs()`

        """
        logger.debug("Generating list of jobs")

        # Create level B inputs
        detectors = self.inputs.detectors
        sampler = self.inputs.sampler
        webdir = self.inputs.webdir
        email = self.inputs.email
        existing_dir = self.inputs.existing_dir

        detectors_list = []
        detectors_list.append(self.inputs.detectors)
        if self.inputs.coherence_test:
            for detector in self.inputs.detectors:
                detectors_list.append([detector])
        sampler_list = self.inputs.sampler
        level_B_prod_list = list(itertools.product(detectors_list, sampler_list))

        level_A_jobs_numbers = range(self.inputs.n_level_A_jobs)
        jobs_inputs = []
        for idx in list(level_A_jobs_numbers):
            for detectors, sampler in level_B_prod_list:
                jobs_inputs.append(
                    JobInput(idx=idx, meta_label=self.inputs.level_A_labels[idx],
                             kwargs=dict(detectors=detectors, sampler=sampler,
                                         webdir=webdir, email=email,
                                         existing_dir=existing_dir)))

        logger.debug("List of job inputs = {}".format(jobs_inputs))
        return jobs_inputs

    def create_summary_jobs(self):
        """ Generate job to generate summary pages """
        for job_input in self.summary_jobs_inputs:
            self.summary_jobs.append(
                self._create_summary_job(job_input))

    def _create_summary_job(self, job_input):
        """ Create a condor job and add it to the dag """
        webdir = job_input.kwargs['webdir']
        email = job_input.kwargs['email']
        detectors = job_input.kwargs["detectors"]
        sampler = job_input.kwargs["sampler"]
        existing_dir = job_input.kwargs['existing_dir']
        idx = job_input.idx
        result_file = '_'.join([self.inputs.label, ''.join(detectors), sampler,
                                job_input.meta_label, "result"])
        job_name = '_'.join([self.inputs.label, 'results_page', str(idx)])
        if job_input.meta_label is not None:
            job_name = '_'.join([job_name, job_input.meta_label])
        job_logs_base = os.path.join(self.inputs.summary_log_directory, job_name)
        submit = self.inputs.submit_directory
        extra_lines = ''
        for arg in ['error', 'log', 'output']:
            extra_lines += '\n{} = {}_$(Cluster)_$(Process).{}'.format(
                arg, job_logs_base, arg[:3])
        extra_lines += '\naccounting_group = {}'.format(self.inputs.accounting)
        extra_lines += '\nx509userproxy = {}'.format(self.inputs.x509userproxy)
        arguments = ArgumentsString()
        arguments.add("webdir", webdir)
        arguments.add("email", email)
        arguments.add("config", self.inputs.ini)
        arguments.add("samples",
                      "{}/{}.h5".format(self.inputs.result_directory,
                                        result_file))
        if existing_dir is not None:
            arguments.add("existing_webdir", existing_dir)

        job = pycondor.Job(
            name=job_name,
            executable=self.summary_executable,
            submit=submit,
            request_memory=self.request_memory, request_disk=self.request_disk,
            request_cpus=self.request_cpus, getenv=self.getenv,
            universe=self.universe, initialdir=self.initialdir,
            notification=self.notification, requirements=self.requirements,
            queue=self.inputs.queue, extra_lines=extra_lines, dag=self.dag,
            arguments=arguments.print(), retry=self.retry, verbose=self.verbose)
        job.add_parent(self.analysis_jobs[idx])
        logger.debug('Adding job: {}'.format(job_name))

    def build_submit(self):
        """ Build the dag, optionally submit them if requested in inputs """
        if self.inputs.submit:
            self.dag.build_submit()
        else:
            self.dag.build()


class ArgumentsString(object):
    """ A convienience object to aid in the creation of argument strings """
    def __init__(self):
        self.argument_list = []

    def append(self, argument):
        self.argument_list.append(argument)

    def add(self, argument, value):
        self.argument_list.append('--{}'.format(argument))
        self.argument_list.append('{}'.format(value))

    def add_unknown_args(self, unknown_args):
        self.argument_list += unknown_args

    def add_command_line_arguments(self):
        """ Adds command line arguments given in addition to the ini file """
        command_line_args_list = utils.get_command_line_arguments()
        # Remove the first positional ini-file argument
        command_line_args_list = command_line_args_list[1:]
        self.argument_list += command_line_args_list

    def print(self):
        return ' '.join(self.argument_list)


class DataDump():
    def __init__(self, label, outdir, trigger_time, interferometers, meta_data,
                 idx):
        self.trigger_time = trigger_time
        self.label = label
        self.outdir = outdir
        self.interferometers = interferometers
        self.meta_data = meta_data
        self.idx = idx

    @property
    def filename(self):
        return os.path.join(
            self.outdir, '_'.join([self.label, str(self.idx), 'data_dump.h5']))

    def to_hdf5(self):
        deepdish.io.save(self.filename, self)

    @classmethod
    def from_hdf5(cls, filename=None):
        """ Loads in a data dump

        Parameters
        ----------
        filename: str
            If given, try to load from this filename

        """
        res = deepdish.io.load(filename)
        if res.__class__ == list:
            res = cls(res)
        if res.__class__ != cls:
            raise TypeError('The loaded object is not a DataDump')
        return res


def main():
    """ Top-level interface for bilby_pipe """
    args, unknown_args = parse_args(
        utils.get_command_line_arguments(), create_parser())
    inputs = MainInput(args, unknown_args)
    # Create a Directed Acyclic Graph (DAG) of the workflow
    Dag(inputs)



#!/usr/bin/env python

import os
import sys
from setuptools import setup
import subprocess

# check that python version is 3.5 or above
python_version = sys.version_info
print('Running Python version %s.%s.%s' % python_version[:3])
if python_version < (3, 5):
    sys.exit('Python < 3.5 is not supported, aborting setup')
else:
    print('Confirmed Python version 3.5.0 or above')


def write_version_file(version):
    """ Writes a file with version information to be used at run time

    Parameters
    ----------
    version: str
        A string containing the current version information

    Returns
    -------
    version_file: str
        A path to the version file

    """
    try:
        git_log = subprocess.check_output(
            ['git', 'log', '-1', '--pretty=%h %ai']).decode('utf-8')
        git_diff = (subprocess.check_output(['git', 'diff', '.']) +
                    subprocess.check_output(
                        ['git', 'diff', '--cached', '.'])).decode('utf-8')
        if git_diff == '':
            git_status = '(CLEAN) ' + git_log
        else:
            git_status = '(UNCLEAN) ' + git_log
    except Exception as e:
        print("Unable to obtain git version information, exception: {}"
              .format(e))
        git_status = ''

    version_file = '.version'
    if os.path.isfile(version_file) is False:
        with open('bilby_pipe/' + version_file, 'w+') as f:
            f.write('{}: {}'.format(version, git_status))

    return version_file


def get_long_description():
    """ Finds the README and reads in the description """
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'README.rst')) as f:
            long_description = f.read()
    return long_description


version = '0.0.3'
version_file = write_version_file(version)
long_description = get_long_description()

setup(name='bilby_pipe',
      description='Automating the running of bilby for gravitational wave signals',
      long_description=long_description,
      url='https://lscsoft.docs.ligo.org/bilby_pipe/index.html',
      author='Gregory Ashton, Isobel Romero-Shaw, Colm Talbot, Charlie Hoy',
      author_email='gregory.ashton@ligo.org',
      license='MIT',
      version=version,
      package_data={'bilby_pipe': [version_file, 'templates/*html']},
      packages=['bilby_pipe'],
      install_requires=[
          'future',
          'pycondor>=0.5',
          'configargparse',
          'ligo-gracedb',
          'gwdatafind',
          'urllib3',
          'deepdish',
          'pesummary'],
      entry_points={
          'console_scripts': [
              'bilby_pipe=bilby_pipe.main:main',
              'bilby_pipe_generation=bilby_pipe.data_generation:main',
              'bilby_pipe_analysis=bilby_pipe.data_analysis:main',
              'bilby_pipe_singularity=bilby_pipe.singularity:main',
              'bilby_pipe_create_injection_file=bilby_pipe.create_injections:main']},
      classifiers=[
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "License :: OSI Approved :: MIT License",
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX'])



#!/usr/bin/env python
"""
Module containing the tools for creating injection files
"""
from __future__ import division, print_function

import sys

import bilby
from bilby.gw.prior import BBHPriorDict
import deepdish
import pandas as pd

from .input import Input
from .utils import parse_args, logger, BilbyPipeError
from .bilbyargparser import BilbyArgParser


def create_parser():
    """ Generate a parser for the create_injections.py script

    Additional options can be added to the returned parser beforing calling
    `parser.parse_args` to generate the arguments`

    Returns
    -------
    parser: BilbyArgParser
        A parser with all the default options already added

    """
    parser = BilbyArgParser(ignore_unknown_config_file_keys=True)
    parser.add('ini', type=str, is_config_file=True, help='The ini file')
    parser.add('--label', type=str, default='LABEL',
               help='The output label')
    parser.add('--outdir', type=str, default='bilby_outdir',
               help='The output directory')
    parser.add_arg('--prior-file', type=str, default=None,
                   help='The prior file from which to generate injections')
    parser.add_arg('--n-injection', type=int,
                   help='The number of injections to generate')
    return parser


class CreateInjectionInput(Input):
    """ An object to hold all the inputs to create_injection

    Parameters
    ----------
    parser: configargparse.ArgParser, optional
        The parser containing the command line / ini file inputs
    args_list: list, optional
        A list of the arguments to parse. Defauts to `sys.argv[1:]`

    """

    def __init__(self, args, unknown_args):
        logger.debug('Creating new CreateInjectionInput object')
        logger.info('Command line arguments: {}'.format(args))

        self.prior_file = args.prior_file
        self.n_injection = args.n_injection
        self.outdir = args.outdir
        self.label = args.label

    @property
    def n_injection(self):
        """ The number of injections to create """
        if self._n_injection is not None:
            return self._n_injection
        else:
            raise BilbyPipeError("The number of injection has not been set")

    @n_injection.setter
    def n_injection(self, n_injection):
        self._n_injection = n_injection

    @property
    def prior_file(self):
        return self._prior_file

    @prior_file.setter
    def prior_file(self, prior_file):
        if isinstance(prior_file, str):
            logger.debug("Setting prior_file to {}".format(prior_file))
            self._prior_file = prior_file
            logger.debug("Generating PriorDict")
            self.prior = BBHPriorDict(filename=self._prior_file)
        else:
            raise BilbyPipeError("Type of prior_file must be str")

    @property
    def prior(self):
        return self._prior

    @prior.setter
    def prior(self, prior):
        if isinstance(prior, bilby.core.prior.PriorDict):
            self._prior = prior
        else:
            raise BilbyPipeError("Input prior not understood")

    def create_injection_file(self, filename):
        logger.info("Generating injection file with prior={}, n_injection={}"
                    .format(self.prior, self.n_injection))
        injection_values = pd.DataFrame.from_dict(self.prior.sample(self.n_injection))
        injections = dict(injections=injection_values, prior=self.prior)
        deepdish.io.save(filename, injections)
        logger.info("Created injection file {}".format(filename))


def main():
    args, unknown_args = parse_args(sys.argv[1:], create_parser())
    inputs = CreateInjectionInput(args, unknown_args)
    inputs.create_injection_file()



"""
A set of generic utilities used in bilby_pipe
"""
import os
import sys
import logging


class BilbyPipeError(Exception):
    def __init__(self, message):
        super().__init__(message)


def get_command_line_arguments():
    """ Helper function to return the list of command line arguments """
    return sys.argv[1:]


def parse_args(input_args, parser):
    """ Parse an argument list using parser generated by create_parser()

    Parameters
    ----------
    input_args: list
        A list of arguments

    Returns
    -------
    args: argparse.Namespace
        A simple object storing the input arguments
    unknown_args: list
        A list of any arguments in `input_args` unknown by the parser

    """

    if len(input_args) == 0:
        raise BilbyPipeError("No command line arguments provided")

    args, unknown_args = parser.parse_known_args(input_args)
    return args, unknown_args


def check_directory_exists_and_if_not_mkdir(directory):
    """ Checks if the given directory exists and creates it if it does not exist

    Parameters
    ----------
    directory: str
        Name of the directory

    """
    if not os.path.exists(directory):
        os.makedirs(directory)
        logger.debug('Making directory {}'.format(directory))
    else:
        logger.debug('Directory {} exists'.format(directory))


def setup_logger(outdir=None, label=None, log_level='INFO',
                 print_version=False):
    """ Setup logging output: call at the start of the script to use

    Parameters
    ----------
    outdir, label: str
        If supplied, write the logging output to outdir/label.log
    log_level: str, optional
        ['debug', 'info', 'warning']
        Either a string from the list above, or an integer as specified
        in https://docs.python.org/2/library/logging.html#logging-levels
    print_version: bool
        If true, print version information
    """

    if '-v' in sys.argv:
        log_level = 'DEBUG'

    if type(log_level) is str:
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError:
            raise ValueError('log_level {} not understood'.format(log_level))
    else:
        level = int(log_level)

    logger = logging.getLogger('bilby_pipe')
    logger.propagate = False
    logger.setLevel(level)

    streams = [type(h) == logging.StreamHandler for h in logger.handlers]
    if len(streams) == 0 or not all(streams):
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)s %(levelname)-8s: %(message)s',
            datefmt='%H:%M'))
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if any([type(h) == logging.FileHandler for h in logger.handlers]) is False:
        if label:
            if outdir:
                check_directory_exists_and_if_not_mkdir(outdir)
            else:
                outdir = '.'
            log_file = '{}/{}.log'.format(outdir, label)
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(logging.Formatter(
                '%(asctime)s %(levelname)-8s: %(message)s', datefmt='%H:%M'))

            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)

    if print_version:
        version = get_version_information()
        logger.info('Running bilby_pipe version: {}'.format(version))


def get_version_information():
    version_file = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), 'bilby_pipe/.version')
    try:
        with open(version_file, 'r') as f:
            return f.readline().rstrip()
    except EnvironmentError:
        print("No version information file '.version' found")


setup_logger(print_version=True)
logger = logging.getLogger('bilby_pipe')
